
(* for emitting timed property if -jps passed (Jean-Paul Special) *)
val JPS = ref ([] : string list);

(* prints the TTS C code by Prn *)
structure TtsCEmit = struct

(* SharedCEmit brings
references;
    xtr  (for hippo only)
    TYPES,TEMP,IMPURE,PURE
functions:
    newline
    getBoundType
    new
    print_Type
    print_ExternDecl
    print_compareFunc
    print_sprintFunc
    print_VarDecl
    assigns
    and assignse
    and print_Exp
    print_Statement
    print_FunctionDecl
    print_Initializer
    print_preFunc
*)
open SharedCEmit;

(* printing header file *)
fun print_Headers (Program ((_,dl,_)::_)) =
    let fun print_dec (TypeDecl (a,ty)) = (print "\n\n"; print_Type ty a; print ";")
	  | print_dec _ = ()
    in newline 0; print "/* fiacre headers */";
       newline 0; print "#include \"fiacre.h\"";
       List.app print_dec dl; newline 0; newline 0
    end;
		

(* Store if a type is shared or not. Used as a cache. *)
val SHARED = ref ([] : string list);
fun shared n = mem n (!SHARED);


local (* to print_pacKFunc *)

(* Returns the smallest C type that can store len elements *)
fun cTypeForLength len = if len <= 255 then
			 	UnsignedCharType
			 else if len <= 65535 then
				UnsignedShortType
			 else
				NatType

(* Returns the equivalent of the C sizeof() operator *)
fun sizeOf BoolType           = 1 (* Converted to a C unsigned char *)
|   sizeOf NatType            = 4
|   sizeOf IntType            = 4
|   sizeOf CharType           = 1
|   sizeOf UnsignedCharType   = 1
|   sizeOf ShortType          = 2
|   sizeOf UnsignedShortType  = 2
|   sizeOf (UnionType cs) =
    let val max = foldr (fn ((_, SOME ty), max) => Int.max (sizeOf ty,max) | (_, max) => max) 0 cs
    in max + sizeOf (cTypeForLength (length cs))
    end
|   sizeOf (EnumType names)      = sizeOf (cTypeForLength (List.length names))
|   sizeOf (QueueType (size,ty)) = sizeOf ty * size + (sizeOf (cTypeForLength (size)))
|   sizeOf (RecordType types)    = foldr (fn ((_,ty),acc) => sizeOf ty + acc) 0 types
|   sizeOf (ArrayType (size,ty)) = sizeOf ty * size
|   sizeOf (NamedType tyName) = sizeOf (getBoundType tyName);

(* Returns the equivalent of the C sizeof() operator, but taking into account that parts of values can be factorized *)
(* We consider that a type is factorized if its size is greater than the sum of
        - The size of the pointer needed for the indirection in the unicity table
        - The size of a cell in the AVL that stores the factorized data: two pointers + height (stored as an unsigned char)
          (it's very conservative: it's for the worst case where data can never be factorized)
   To be conservative, we assume that pointers are 64 bits.
   Thus, the final minimal size required is:
        1 * pointer + 2 * pointer + unsigned char

   Furthermore, we need to add the size of the type chosen to encode the length of the compressed factorized data. Again, to be conservative,
   we consider the maximum encoding type (e.g. unsigned int)
*)

val ptrSize = 8;
val factorizationThreshold = 3 * ptrSize + 1 + sizeOf NatType;

in

(* determines if a datatype is shared or not *)
(* never applied to a named type *)
(* "shared" check is done "bottom up", so named types found have been checked yet *)
fun shareable ty =
    (case ty of
	 RecordType [_] => false
       | RecordType types =>
	 foldr (fn ((_,ty),acc) => shSizeOf ty + acc) 0 types >= factorizationThreshold
       | ArrayType (size,nty) =>
	 if case nty of BoolType => size <= 32 | _ => false then
	     false
	 else
	     shSizeOf nty * size >= factorizationThreshold
       |  _ => false)
and shSizeOf ty =
	    case ty of
		BoolType            => 1 (* Converted to a C unsigned char *)
	      | NatType             => 4
	      | IntType             => 4
	      | CharType            => 1
	      | UnsignedCharType    => 1
	      | ShortType           => 2
	      | UnsignedShortType   => 2
	      | UnionType cs        =>
		let val max = foldr (fn ((_, SOME ty), max) => Int.max(shSizeOf ty,max) |   (_, max) => max) 0 cs
		in max + (shSizeOf (cTypeForLength (length cs)))
		end
	      | EnumType names      => shSizeOf (cTypeForLength (List.length names))
	      | QueueType (size,ty) => (* queues no more shared *)
		shSizeOf ty * size + (shSizeOf (cTypeForLength (size)))
	      | RecordType [(_, ty)] => shSizeOf ty
	      | RecordType types     => foldr (fn ((_,ty),acc) => shSizeOf ty + acc) 0 types
	      | ArrayType (size,ty) =>
		if case ty of BoolType => size <= 32 | _ => false then
		    if size <= 8  then
			shSizeOf UnsignedCharType
		    else if size <= 16 then
			shSizeOf UnsignedShortType
		    else
			shSizeOf NatType
		else shSizeOf ty * size
	      | NamedType tyName    => 
		if shared tyName then ptrSize else shSizeOf (getBoundType tyName);

(* assuming SHARED has been computed: *)
fun sharedSizeOf ty = if !SHARING then shSizeOf ty else sizeOf ty;

fun print_packFunc (typeName, ty) =
let
	val isShared = shared typeName
	val size = (sharedSizeOf ty) * 2 (* worst-case scenario *)
	val encodingTmp = cTypeForLength size
	val sharedLenEncoding = case cTypeForLength (size + (sizeOf encodingTmp)) of
		            	  UnsignedCharType  => "unsigned char"
		          	| UnsignedShortType => "unsigned short"
		          	| NatType           => "unsigned int"
	
	fun pack container varName (NamedType typeName) = 
	let
		val address = case getBoundType typeName of
				CharType          => ""
			      |	UnsignedCharType  => ""
			      |	ShortType         => ""	
			      | UnsignedShortType => ""
			      | IntType           => ""
			      | NatType  	  => ""
                              | BoolType          => ""
			      | EnumType _        => ""
			      | _                 => "&"
	in
	    if shared typeName then
		print ("pack_shared_" ^ typeName ^ "(" ^ address ^ container ^ varName ^ ", data)")
	    else
		print ("pack_" ^ typeName ^ "(" ^ address ^ container ^ varName ^ ", data)")
	end
	|   pack container name BoolType = print ("pack_bool(" ^ container ^ name ^ ", data)")
	|   pack container name CharType = print ("pack_char(" ^ container ^ name ^ ", data)")
	|   pack container name UnsignedCharType = print ("pack_unsigned_char(" ^ container ^ name ^ ", data)")
	|   pack container name ShortType = print ("pack_short(" ^ container ^ name ^ ", data)")
	|   pack container name UnsignedShortType = print ("pack_unsigned_short(" ^ container ^ name ^ ", data)")
	|   pack container name IntType = print ("pack_int(" ^ container ^ name ^ ", data)")
	|   pack container name NatType = print ("pack_unsigned_int(" ^ container ^ name ^ ", data)")

	fun unpack container varName (NamedType typeName) =
	    if shared typeName then
		print ("unpack_shared_" ^ typeName ^ "(&(" ^ container ^ varName ^ "), data)")
	    else
		print ("unpack_" ^ typeName ^ "(&(" ^ container ^ varName ^ "), data)")
	|   unpack container name BoolType = print ("unpack_bool(&(" ^ container ^ name ^ "), data)")
	|   unpack container name CharType = print ("unpack_char(&(" ^ container ^ name ^ "), data)")
	|   unpack container name UnsignedCharType = print ("unpack_unsigned_char((unsigned char*)&(" ^ container ^ name ^ "), data)")
	|   unpack container name ShortType = print ("unpack_short(&(" ^ container ^ name ^ "), data)")
	|   unpack container name UnsignedShortType = print ("unpack_unsigned_short((unsigned short*)&(" ^ container ^ name ^ "), data)")
	|   unpack container name IntType = print ("unpack_int(&(" ^ container ^ name ^ "), data)")
	|   unpack container name NatType = print ("unpack_unsigned_int(&(" ^ container ^ name ^ "), data)")

	fun printSharedFunsStr lenEnc =
	let
		val storageName = typeName ^ "_storage"
	in
		newline 0;
		newline 0; print "#ifdef SHARE";
		newline 0; print ("void pack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* x, char* data) {");
		newline 2; print (lenEnc ^ " size = sizeof(" ^ lenEnc ^ ");");
		newline 2; print ("size += pack_" ^ typeName ^ "(x,data + size);");
		newline 2; print ("*(" ^ lenEnc ^ " *)data = size;");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("void unpack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* x, char* data) {");
		newline 2; print ("unpack_" ^ typeName ^ "(x,data + sizeof(" ^ lenEnc ^ "));");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("size_t size_" ^ typeName ^ " (char* data) {");
		newline 2; print ("return *(" ^ lenEnc ^ " *)data;");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("#define pack_shared_" ^ typeName ^ "(x,data)   pack_key(&" ^ typeName ^ "_storage,x,data)");
		newline 0; print ("#define unpack_shared_" ^ typeName ^ "(x,data) unpack_key(&" ^ typeName ^ "_storage,x,data)");
		newline 0;
		newline 0; print "#else";
		newline 0; print ("#define pack_shared_" ^ typeName ^ "(x,data)   pack_" ^ typeName ^ "(x,data)");
		newline 0; print ("#define unpack_shared_" ^ typeName ^ "(x,data) unpack_" ^ typeName ^ "(x,data)");
		newline 0; print "#endif"
	end

in
	case String.sub(typeName,0) of
		#"e" =>
		let
			val EnumType elems = ty
			val encoding = cTypeForLength (length elems)
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (enum " ^ typeName ^ " e, char* data) {");
			newline 0; print "  return "; pack "" "e" encoding; print ";";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (enum " ^ typeName ^ "* e, char* data) {");
			newline 0; print "  unsigned char tmp;";
			newline 0; print "  const size_t res = "; unpack "" "tmp" encoding; print ";";
			newline 0; print "  *e = tmp;";
			newline 0; print "  return res;";
			newline 0; print "}"
		end

	|	#"u" =>
		let
			val UnionType cs = ty
			val lenEncoding = cTypeForLength (List.length cs)
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* u, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  data += "; pack "u->" "con" lenEncoding; print ";";
			newline 0; print "  switch (u->con) {";
			List.app (fn (n, SOME ty) =>  (newline 4;print ("case " ^ n ^ ": data += "); pack  "u->arg." n ty; print "; break;")
				 |   (_,NONE) => ()
				 )
				 cs;
			newline 0; print "    default: break;";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* u, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  unsigned char tmp;";
 			newline 0; print "  data += "; unpack "" "tmp" lenEncoding; print ";";
			newline 0; print "  u->con = tmp;";
			newline 0; print "  switch (u->con) {";
			List.app (fn (n, SOME ty) => (newline 4;print ("case " ^ n ^ ": data += "); unpack  "u->arg." n ty; print "; break;")
				 |    (_, NONE) => ()
				 )
				 cs;
			newline 0; print "    default: break;";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}"
		end

	|	#"q" =>
		let
			val QueueType(size,eltType) = ty
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* q, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  int i = 0;";
			newline 0; print "  data += pack_unsigned_int(q->len, data);";
			newline 0; print "  for( i = 0; i < q->len; ++i) {";
			newline 0; print "    data += "; pack "q->" "at[i]" eltType; print";";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* q, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  int i = 0;";
			newline 0; print "  data += unpack_unsigned_int(&q->len, data);";
			newline 0; print "  for( i = 0; i < q->len; ++i) {";
			newline 0; print "    data += "; unpack "q->" "at[i]" eltType; print";";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}"
		end

	|	#"a" =>
		let
			val ArrayType(size,eltType) = ty
		in
			if case eltType of BoolType => size <= 32 | _ => false then
			let
			    val enc = if size <= 8 then "unsigned char" else if size <= 16 then "unsigned short" else "unsigned int"
			in 
			    newline 0;
			    newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
			    newline 0; print ("  size_t i; " ^ enc ^ " t = 0;");
			    newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
			    newline 0; print  "    t |= a->at[i] << i;";
			    newline 0; print  "  }";
			    newline 0; print ("  memcpy(data, &t, sizeof(" ^ enc ^ "));");
			    newline 0; print ("  return sizeof(" ^ enc ^ ");");
			    newline 0; print  "}";
			    newline 0;
			    newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
			    newline 0; print ("  size_t i; " ^ enc ^ " t = 0;");
			    newline 0; print ("  memcpy(&t, data, sizeof(" ^ enc ^ "));");
			    newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
			    newline 0; print  "    a->at[i] = t & 0x01;";
			    newline 0; print  "    t >>= 1;";
			    newline 0; print  "  }";
			    newline 0; print ("  return sizeof(" ^ enc ^ ");");
			    newline 0; print  "}"
			end
			else
			(	newline 0;
				newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
				newline 0; print  "  const char* orig = data;";
				newline 0; print  "  size_t i = 0;";
				newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
				newline 0; print  "    data += "; pack "a->" "at[i]" eltType; print ";";
				newline 0; print  "  }";
				newline 0; print "  return data - orig;";
				newline 0; print  "}";
				newline 0;
				newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
				newline 0; print  "  const char* orig = data;";
				newline 0; print  "  size_t i = 0;";
				newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
				newline 0; print  "    data += "; unpack "a->" "at[i]" eltType; print ";";
				newline 0; print  "  }";
				newline 0; print  "  return data - orig;";
				newline 0; print  "}";
				if isShared then
				    printSharedFunsStr sharedLenEncoding
				else
				    ()
			)
		end

	|       #"r" =>
		let
			val RecordType types = ty
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* r, char* data) {");
			newline 0; print  "  const char* orig = data;";
			List.app (fn (name,ty) => (newline 2; print "data += "; pack "r->" name ty; print ";")) types;
			newline 0; print  "  return data - orig;";
			newline 0; print  "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* r, char* data) {");
			newline 0; print  "  const char* orig = data;";
			List.app (fn (name,ty) => (newline 2; print "data += "; unpack "r->" name ty; print ";")) types;
			newline 0; print  "  return data - orig;";
			newline 0; print  "}";
			if isShared then
			    printSharedFunsStr sharedLenEncoding
			else
			    ()
		end

	|       #"v" =>
		let
		    val (shNeedHash,shLenEncoding) =
			let val tySize = sharedSizeOf ty * 2 (* *2 => worst-case compression scenario*)
			    val needHash = tySize > 255
			    val size = if needHash then tySize + 4 (* 4 bytes to store an hash value *) else tySize
			    val encodingTmp = cTypeForLength size
			in (tySize > 256,
			    case cTypeForLength (size + (sizeOf encodingTmp)) of
				UnsignedCharType  => "unsigned char"
			      | UnsignedShortType => "unsigned short"
			      | NatType           => "unsigned int")
			end
		    val (needHash,lenEncoding) =
			let val tySize = sizeOf ty * 2 (* *2 => worst-case compression scenario*)
			    val needHash = tySize > 256
			    val size = if needHash then tySize + 4 (* 4 bytes to store an hash value *) else tySize
			    val encodingTmp = cTypeForLength size
			in (tySize > 256,
			    case cTypeForLength (size + (sizeOf encodingTmp)) of
				UnsignedCharType  => "unsigned char"
			      | UnsignedShortType => "unsigned short"
			      | NatType           => "unsigned int")
			end
		    val RecordType types = ty
		    val someshared = (* avoid SHARE if no change in size/hash whether SHARE of not *)
			not (null (!SHARED)) andalso (lenEncoding <> shLenEncoding orelse needHash <> shNeedHash)
		in
			(* values always stored with their size and possibly hash value, but size depends on SHARE *)

		        newline 0;
			newline 0; print ("size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			newline 0; print ("  const char* orig = data;");
			List.app (fn (name,ty) => (newline 2; print "data += "; pack "v->" name ty; print ";")) types;
			newline 0; print ("  return data - orig;");
			newline 0; print "}";
			newline 0;

			newline 0; print ("size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			newline 0; print ("  const char* orig = data;");
			List.app (fn (name,ty) => (newline 2; print "data += "; unpack "v->" name ty; print ";")) types;
			newline 0; print ("  return data - orig;");
			newline 0; print "}";
			newline 0;

			newline 0;
			newline 0; print ("void pack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			let fun p () =
				if needHash then
				    (newline 0; print ("  " ^ lenEncoding ^ " size = sizeof(int) + sizeof(" ^ lenEncoding ^ ");"))
				else
				    (newline 0; print ("  " ^ lenEncoding ^ " size = sizeof(" ^ lenEncoding ^ ");"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  " ^ shLenEncoding ^ " size = sizeof(int) + sizeof(" ^ shLenEncoding ^ ");"))
				else
				    (newline 0; print ("  " ^ shLenEncoding ^ " size = sizeof(" ^ shLenEncoding ^ ");"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "  size += pack_value(v,data + size);";
			let fun p () =
				if needHash then
				    (newline 0; print ("  *(" ^ lenEncoding ^ " *)(data + sizeof(int)) = size;");
				     newline 0; print ("  MurmurHash3_x86_32(data + sizeof(int), size - sizeof(int), 42, (void*)data);"))
				else
				    (newline 0; print ("  *(" ^ lenEncoding ^ " *)data = size;"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  *(" ^ shLenEncoding ^ " *)(data + sizeof(int)) = size;");
				     newline 0; print ("  MurmurHash3_x86_32(data + sizeof(int), size - sizeof(int), 42, (void*)data);"))
				else
				    (newline 0; print ("  *(" ^ shLenEncoding ^ " *)data = size;"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}";

			newline 0;
			newline 0; print ("void unpack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			let fun p () =
				if needHash then
				    (newline 0; print ("  unpack_value(v,data + sizeof(int) + sizeof(" ^ lenEncoding ^ "));"))
				else
				    (newline 0; print ("  unpack_value(v,data + sizeof(" ^ lenEncoding ^ "));"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  unpack_value(v,data + sizeof(int) + sizeof(" ^ shLenEncoding ^ "));"))
				else
				    (newline 0; print ("  unpack_value(v,data + sizeof(" ^ shLenEncoding ^ "));"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}";

			newline 0;
			newline 0; print "size_t size_value(char* data) {";
			let fun p () =
				if needHash then
				    (newline 0; print ("  return *(" ^ lenEncoding ^ " *)(data + sizeof(int));"))
				else
				    (newline 0; print ("  return *(" ^ lenEncoding ^ " *)(data);"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  return *(" ^ shLenEncoding ^ " *)(data + sizeof(int));"))
				else
				    (newline 0; print ("  return *(" ^ shLenEncoding ^ " *)(data);"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}"
				   
		end
end;

end; (* local *)

fun sprint_value_tts spr ftyl =
    (newline 0; print "\nstatic int sprint_value (int sz, char *buff, key s, const struct pmap *pmap) {";
     newline 2; print "struct value *v = lookup(s);";
     newline 2; print "avail = sz - 1;";
     newline 2; print "/* prints state into buff: */";
     newline 2; print "PUTS(pmap->r.l);";
     spr ftyl;
     newline 2; print "PUTS(pmap->r.r);";
     newline 2; print "return (avail<0) ? -1 : (sz - avail - 1);";
     newline 0; print "}";
     newline 0; print "\nint sprint_state (int sz, char *buff, key s) {";
     newline 2; print "return sprint_value (sz, buff, s, &fcrPmap);";
     newline 0; print "}";
     newline 0; print "\nint sprint_json_state (int sz, char *buff, key s) {";
     newline 2; print "return sprint_value (sz, buff, s, &jsonPmap);";
     newline 0; print "}");    
	  
(* emit type declarations for named types *)
fun print_TypeDecl (TypeDecl (a,ty)) =
    ((* print "\n\n"; print_Type ty a; print ";" *)  (* now printed in header file *)
     print "\n\n/* utilities for structure "; print a; print " */";
     print_compareFunc (a,ty);
     print_sprintFunc sprint_value_tts (a,ty);
     print_packFunc (a,ty));

fun print_initialFunc a b = (* a = state variables, b = rtmp variables *)
    (newline 0; newline 0; print "/* initializes temporaries and creates initial state */";
     newline 0; print "key initial() {";
     newline 2; print "struct value start;";
     (* initialize temporaries: *)
     if null b then () else (newline 2; print "/* read-only temporaries: */");
     List.app (fn x as (st,_,v) => print_Initializer st v "") b;
     (* initialize state variables: *)
     newline 2; print "/* state variables: */";
     List.app (fn x as (st,_,v) => print_Initializer st v "start.") a;
     (* establishes value probes for initial state *)
     if !PROPERTIES andalso not (null (!VALUEVARS)) then
	 (* List.app (fn (_,v,e,s) => (newline 2; print "start."; print v; print " = ";   print_Exp e "start."; print ";")) (!VALUEVARS) *)
	 (newline 2; print "/* initializing value probes */";
	  List.app (fn (_,_,_,ref s) => print_Statement 2 s "start.") (!VALUEVARS))
     else
	 ();
     newline 2; print "/* storage initializations: */";
     newline 2; print "open_storage();";
     newline 2; print "values_storage = init_storage(sizeof(struct value), compare_value, pack_with_size_value, unpack_with_size_value, size_value);";
     case !SHARED of
	 [] => ()
       | sh => (newline 0; print "#ifdef SHARE";
		List.app (fn name =>
			     (newline 2;
			      print (name ^ "_storage = init_storage(sizeof(struct " ^ name ^ ")"
				     ^                        ", compare_" ^ name
				     ^                        ", pack_with_size_" ^ name
				     ^                        ", unpack_with_size_" ^ name
				     ^                        ", size_" ^ name
				     ^                        ");")))
			 sh;
		newline 0; print "#endif");
     newline 2; print "return store(&start);"; 
     newline 2; print "error: return 0;"; 
     newline 0; print "}";
     newline 0;
     newline 0; print "int final(int debug) {";
     newline 2; print ("finalize_storage(\"top\", debug, &values_storage);");
     case !SHARED of
	 [] => ()
       | sh => (newline 0; print "#ifdef SHARE";
		List.app (fn name => (newline 2; print ("finalize_storage(\"" ^ name ^ " \", debug, &" ^ name ^ "_storage);"))) sh;
		newline 0; print "#endif");
     newline 0; print "  return close_storage(debug);";
     newline 0; print "}");

fun print_pre_tts print_guard e tag index dl k =
    (newline 0; newline 0; print ("bool pre_"^ Int.toString(index) ^" (key s) {");
     newline 2; print "/* "; print tag; print " */";
     (* if null k then () else (newline 2; print "/* temporaries */"); *)
     List.app (fn d => (newline 2; print_VarDecl d)) k;
     if assignse dl e
     then (* #### improve: make local copies of state variables updated in the block, rather than all #### *)
	 (newline 2; print "struct value new = *(struct value *)(lookup(s));";
	  print_guard dl e "new.")
     else (newline 2; print ("struct value *v = lookup(s);");
	   print_guard dl e "v->");
     newline 2; print "return 1;"; 
     newline 1; print ("error: return 0;");
     newline 0; print "}");
    
fun print_actFunc tag index stm k =
    if assigns [] stm
    then let fun hasGuard s = case s of StGuard _ => true | StBlock (_,s) => hasGuard s | _ => false
	     fun print_act s dl =
		 case s of
		     StBlock(dl',s) => print_act s (dl @ dl')
		   | StSequence(s1,s2) =>
		     if matchStm hasGuard s1
		     then print_act s2 dl
		     else (newline 0; newline 0; print ("key act_"^ Int.toString(index) ^" (key s) {");
			   newline 2; print "/* "; print tag; print " */";
			   (* if null k then () else (newline 2; print "/* temporaries */"); *)
			   List.app (fn d => (newline 2; print_VarDecl d)) k;
			   newline 2; print ("struct value new = *(struct value*)lookup(s);");
			   print_Statement 2 s "new.";
			   newline 2; print ("return store(&new);");
			   newline 1; print ("error: return s;");
			   newline 0; print "}")
		   | _ => ()
	 in print_act stm []
	 end
    else ();

fun print_transFunc a rk =
    (newline 0; print "\n/* -- predicates and actions */";
     newline 0; print "/* by default, pre_t* returns true and act_t* is the identity function */";
     newline 0; newline 0; print "/* translation table for transition names */";
     newline 0; print ("char *transtable["^ Int.toString(List.length a) ^"] = "); 
     print_list ("{",",","}") (fn (([st],_,_),_,_,_,_) => print ("\""^ hd (String.fields (fn #":" => true | _ => false) st)^"\"")) a; print ";";
     newline 0; print "\nchar **transitions (int *sz) {";
     newline 0; print "*sz ="; print (Int.toString(List.length a)); print ";";
     newline 0; print "return transtable;";
     newline 0; print "}";
     List.app (fn (i,(([tag],lab,_),_,stm,_,_)) => print_preFunc print_pre_tts (tag^(case lab of "" => "" | _ => (":"^lab))) i stm [] rk)
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a));
     List.app (fn (i,(([tag],lab,sec::tags),_,stm,_,_)) =>
		  (if !TAGS andalso not (null tags) then
		       let open TextIO
			   fun p s = output (stdOut,s)
		       in p tag; p " ->"; List.app (fn g => (p " "; p g)) tags; p "\n"
		       end
		   else
		       ();
		   print_actFunc (tag^(case lab of "" => "" | _ => (":"^lab))) i stm rk))
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a)));

(* prints independent function *)
fun print_indepFunc varL =
    (newline 0; newline 0; print "/* -- independance relation for transitions, for partial order modes and reset in timed modes -- */";
     newline 0; print "/* indices refer to transtable; specifies independance for ALL transitions */";
     newline 0; print "bool independant (int t1, int t2) {";
     newline 2; print "  /* always true */";
     newline 2; print "  return 1;";
     newline 0; print "}");

(* prints structure components in C style *)
fun print_obsFunc varL =
    let	fun getVarNames st ty =
	    case ty of
		BoolType => [("(int)(v->"^st^")",st)]
	      | NatType => [("(int)(v->"^st^")",st)]
	      | IntType => [("(int)(v->"^st^")",st)]
	      | CharType => [("(int)(v->"^st^")",st)]
	      | UnsignedCharType => [("(int)(v->"^st^")",st)]
	      | ShortType => [("(int)(v->"^st^")",st)]
	      | UnsignedShortType => [("(int)(v->"^st^")",st)]
	      | EnumType _ => [("(int)(v->"^st^")",st)]
	      | UnionType cs => (* [("(int)(v->"^st^".con)",st)] *)
		let val con as (c,_) = ("(int)(v->"^st^".con)",st^".con")
		in con :: List.concat (List.map (fn (n,SOME ty) =>
						    let val ly = getVarNames (st^".arg."^n) ty
						    in map (fn (uu,vv) => ("("^c^" == "^n^") ? ("^uu^") : 0",vv)) ly
						    end
						  | (_,NONE) => []) cs)
		end
	      | RecordType stList => List.concat(List.map (fn (x,y) => getVarNames (st^"."^x) y) stList)
	      | ArrayType(s,tty) => List.concat(List.tabulate(s,fn y => getVarNames (st^".at["^ Int.toString(y)  ^"]") tty))
	      | QueueType (s,tty) =>
		let val len as (c,_) = ("(int)(v->"^st^".len)",st^".len")
		in len :: List.concat(List.tabulate(s,fn y => let val ly = getVarNames (st^".at["^ Int.toString(y)  ^"]") tty
							      in map (fn (uu,vv) => ("("^c^" > "^Int.toString y^") ? ("^uu^") : 0",vv)) ly
							      end))
		end
	      | NamedType aa => getVarNames st (getBoundType aa)
	val (str1,str2) = ListPair.unzip (flat (map (fn (stl,b,c) => getVarNames stl b) varL))
	fun encode n = (* surrounds n with braces if contains "." or "[" *) (* ### should suppress ".at" too ? ### *)
	    if null (List.filter (fn #"." => true | #"[" => true | _ => false) (String.explode n))
	    then "\"" ^ n ^ "\"" else "\"{" ^ n ^ "}\""
    in  newline 0; print "\n/* -- state observers for kts output -- */";
	newline 0; print ("char *obsnames[" ^ Int.toString(List.length str2) ^ "] = {");
	print_list ("",",","") (fn y => print (encode y)) str2; print "};";
	newline 0; print ("\nchar **obs_names (int *count) {*count = " ^ Int.toString(List.length str2)  ^"; return obsnames;}");
	newline 0; print ("\n/* returns values of observables for state s (abstractions for noninteger fields) */");
	newline 0; print ("int ovalues["^ Int.toString (List.length str2)  ^"];");
	newline 0; newline 0; print "int *obs_values (key s) {";
	newline 0; print "struct value *v = lookup(s);";
	case str2 of
	    [] => ()
	  | _ => (newline 0; print_list ("",";\n",";")
					(fn (x,y) => print ("ovalues[" ^ Int.toString(x) ^ "]=" ^ y))
					(ListPair.zip(List.tabulate(List.length str1,(fn y=>y)),str1)));
	newline 0; print "return ovalues;";
	newline 0; print "}"
    end;

(* printing code file *)
fun print_Fiacre (Program ((_,dl,d)::_)) =
    let fun print_dec d =
	    case d of
		FunctionDecl _ => (newline 0; print_FunctionDecl d)
	      | ExternDecl _ => (newline 0; print_ExternDecl d)
	      | TypeDecl (s,ty) => print_TypeDecl d
	      | _ => ()
    in newline 0; newline 0; print "static int avail;    /* for sprint_* */";
       newline 0; print "static char tmpbuff[16]; /* for sprint_* */";
       List.app print_dec dl;
       let val ProcessDecl (a,_,_,_,rk,k,f,_,_,_,j) = d
       in newline 0; newline 0; print "/* Process "; print a; print " */";
	  TEMP := map (fn (v,_,_) => v) (rk @ k);
	  if null rk then () else (newline 0; newline 0; print "/* read-only temporary variables */");
	  List.app (fn d => (newline 0; print_VarDecl d)) rk;
	  newline 0;
	  (* creates value type, compare_value, sprint_value: *)
	  newline 0; print "/* state type */";
	  let val valuetype = RecordType (map (fn (y,ty,_) => (y,ty)) f)
	  in print "\n\n"; print_Type valuetype "value"; print ";";
	     print_TypeDecl (TypeDecl ("value",valuetype))  (* functions *)
	  end;
	  print_initialFunc f rk;
	  print_transFunc j k;
	  print_indepFunc f;
	  print_obsFunc f
       end;
       newline 0
    end;

(* emits .c iff some data *)
fun emitFiacre (hfile,fdef,writeh) (p as Program ((_,dl,d as ProcessDecl (_,_,_,_,rk,_,f,_,_,_,_))::_)) =
      (print ("/* Created " ^ Date.toString (Date.fromTimeLocal(Time.now ())) ^ " by frac version " ^ version ^ " */");
       newline 0; newline 0; print "#include <stdlib.h>";
       newline 0; print "#include <math.h>";
       newline 0; print "#include <string.h>";
       newline 0; print "#include <stdio.h>";
       newline 0; print "#include <stddef.h>";
       newline 0; print "\n/* tts API version */";
       newline 0; print "int ttsapiversion () {return 2;}";
       newline 0; print "\n/* provider */";
       newline 0;
       print "char *source_info[3] = {\"frac\",\"";
       print version; print "\",\"";
       print (Date.toString (Date.fromTimeLocal(Time.now ()))); print "\"};";
       print "\nchar **sourceinfo () {return source_info;}";
       flush ();
       (* following is disabled as there may be non trivial guards even if no data
          e.g. due to pattern matching used in a communication (PAB example).
	  disabling it should not harm, in any case *)
       if false (* List.null f andalso List.null rk *) then
	   (newline 0; print "/* minimal functions */";
	    newline 0; print "typedef int bool;";
	    newline 0; print "typedef long key;";
	    newline 0; print "char buff[3];";
	    newline 0; print "int sprint_state (int sz, char *buff, key s) {strcpy(buff,\"{}\\0\"); return 2;}";
	    newline 0; print "int sprint_json_state (int sz, char *buff, key s) {strcpy(buff,\"{}\\0\"); return 2;}";
	    newline 0; print "key initial() {return 0;}";
	    newline 0; print "char **transitions (int *sz) {*sz =0; return 0;}";
	    newline 0; print "char **obs_names (int *count) {*count = 0; return 0;}";
	    newline 0; print "int *obs_values (key s) {return 0;}";
	    newline 0; print "bool independant (int t1, int t2) {return 1;}";
	    newline 0; print "int final(int debug) {return 0;}";
	    newline 0; flush ();
	    OS.FileSys.remove fdef handle _ => ())
       else
	   let val p' as Program ((_,dl',_)::_) = Name.nameTypes p
	   in (* Construct the list of types (in order of declaration) *)
	      TYPES := List.mapPartial (fn (TypeDecl(a,ty)) => SOME (a,ty) | _ => NONE) dl';
	      (* Compute the shared types bottom up (IMPORTANT!) *)
	      if !SHARING then
		  (List.app (fn (name,ty) => if shareable ty then SHARED := name :: !SHARED else ()) (!TYPES);
		   SHARED := rev (!SHARED))  (* not essential, just for code readability *)
	      else
		  ();
	      (* Do not share a type if it's alone at the top-level *)
	      case f of [(_, NamedType tyName, _)] => SHARED := List.filter (fn n => n <> tyName) (!SHARED) | _ => ();
	      writeh print_Headers p';
	      newline 0; print "\n/* fiacre and private headers */";
	      newline 0; print "#include \""; print hfile; print "\"";
	      newline 0; newline 0; print "/* storage headers */";
	      newline 0; print "#include \"pack.h\"";
	      newline 0; print "#include \"store.h\"";
	      newline 0;
	      newline 0; print "static storage_t values_storage;";
	      case !SHARED of
		  [] => ()
		| sh => (newline 0; print "#ifdef SHARE";
			 List.app (fn name => (newline 0; print ("static storage_t " ^ name ^ "_storage;"))) sh;
			 newline 0; print "#endif");
	      newline 0; print "#define lookup(s)   lookup_data(&values_storage,s)";
	      newline 0; print "#define store(v)    store_data(&values_storage,v)";
	      print_Fiacre p';
	      newline 0; flush ()
	   end)

       handle BYE => raise BYE | _ => failwith "emit c";
   
end; (* structure TtsCEmit *)


val ltllib = String.concat
[
"\n", 
"\n# general",
"\nop deadlockFree = [] - dead;", 
"\nop infinitelyOften L = [] <> L;",
"\nop mortal L = <> [] - L;",
"\n",
"\n",
"\n# patterns, source http://patterns.projects.cis.ksu.edu/documentation/patterns.shtml",
"\n",
"\n# Weak until",
"\ninfix 0 p W q = p U (q \\/ []p);",
"\n",
"\n# absence",
"\nop absent P = [] - P;",
"\nop absentBefore P R = <>R => (-P U R);",
"\nop absentAfter P Q = [](Q => [] - P);",
"\nop absentBetween P Q R = [] ((Q /\\ -R /\\ <>R) => (-P U R));",
"\nop absentUntil P Q R = [](Q /\\ -R => (-P W R));",
"\n",
"\n# presence",
"\nop present P = <> P;",
"\nop presentBefore P Q = -Q W (P /\\ -Q);",
"\nop presentAfter P Q = [] - Q \\/ <> (Q /\\ <>P);",
"\nop presentBetween P Q R = [] (Q /\\ -R => (-R W (P /\\ -R)));",
"\nop presentUntil P Q R = [] (Q /\\ -R => (-R U (P /\\ -R)));",
"\n",
"\n# universality",
"\nop always P = [] P;",
"\nop alwaysBefore P R = <> R => (P U R);",
"\nop alwaysAfter P Q = [] (Q => []P);",
"\nop alwaysBetween P Q R = [] ((Q /\\ -R /\\ <>R) => (P U R));",
"\nop alwaysUntil P Q R = [] (Q /\\ -R => (P W R));",
"\n",
"\n# precedence",
"\nop precedes S P = -P W S;",
"\nop precedesBefore S P R = <>R => (-P U (S \\/ R));",
"\nop precedesAfter S P Q = [] - Q \\/ <>(Q /\\ (-P W S));",
"\nop precedesBetween S P Q R = []((Q /\\ -R /\\ <>R) => (-P U (S \\/ R)));",
"\nop precedesUntil S P Q R = [](Q /\\ -R => (-P W (S \\/ R)));",
"\n",
"\n# response",
"\nop leadsto P S = [] (P => <>S);",
"\nop leadstoBefore P S R = <>R => (P => (-R U (S /\\ -R))) U R;",
"\nop leadstoAfter P S Q = [] (Q => [](P => <>S));",
"\nop leadstoBetween P S Q R = [] ((Q /\\ -R /\\ <>R) => (P => (-R U (S /\\ -R))) U R);",
"\nop leadstoUntil P S Q R = [] (Q /\\ -R => ((P => (-R U (S /\\ -R))) W R));",
"\n",
"\n"
];


val mmclib = String.concat
[
"\n", 
"\n# ctl",
"\nprefix AF g = min x | g \\/ (<T>T /\\ [T]x);",
"\nprefix EF g = min x | g \\/ <T>x;",
"\nprefix AG g = max x | g /\\ [T]x;",
"\nprefix EG g = max x | g /\\ <T>x;",
"\ninfix f AU g = min x | g \\/ (f /\\ <T>x /\\ [T]x);",
"\ninfix f EU g = min x | g \\/ (f /\\ <T>x);",
"\n", 
"\n# general",
"\nop deadlockFree = - <T>T;", 
"\nop live L = AG EF L;",
"\nop mortal L = AF EG - L;",
"\n",
"\n",
"\n# patterns, source http://patterns.projects.cis.ksu.edu/documentation/patterns.shtml",
"\n",
"\n# Weak AW",
"\ninfix f AW g = (-g) EU (-f /\\ -g);",
"\n",
"\n# absence",
"\nop absent P = AG - P;",
"\nop absentBefore P R = (-P \\/ AG - R) AW R;",
"\nop absentAfter P Q = AG (Q => AG - P);",
"\nop absentBetween P Q R = AG (Q /\\ -R => ((-P \\/ AG - R) AW R));",
"\nop absentUntil P Q R = AG (Q /\\ -R => (-P AW R));",
"\n",
"\n# presence",
"\nop present P = AF P;",
"\nop presentBefore P R = (- R) AW (P /\\ (- R));",
"\nop presentAfter P Q = -Q AW (Q /\\ AF P);",
"\nop presentBetween P Q R = AG (Q /\\ -R => -R AW (P /\\ -R));",
"\nop presentUntil P Q R =   AG (Q /\\ -R => -R AU (P /\\ -R));",
"\n",
"\n# universality",
"\nop always P = AG P;",
"\nop alwaysBefore P R = (P \\/ AG - R) AW R;",
"\nop alwaysAfter P Q = AG (Q => AG P);",
"\nop alwaysBetween P Q R = AG (Q /\\ -R => ((P \\/ AG - R) AW R));",
"\nop alwaysUntil P Q R = AG (Q /\\ -R => P AW R);",
"\n",
"\n# precedence",
"\nop precedes S P = -P AW S;",
"\nop precedesBefore S P R = (-P \\/ AG - R) AW (S \\/ R);",
"\nop precedesAfter S P Q = -Q AW (Q /\\ (-P AW S));",
"\nop precedesBetween S P Q R = AG (Q /\\ -R => (-P \\/ AG - R) AW (S \\/ R));",
"\nop precedesUntil S P Q R = AG (Q /\\ -R => (-P AW (S \\/ R)));",
"\n",
"\n# response",
"\nop leadsto P S = AG (P => AF S);",
"\nop leadstoBefore P S R = ((P => -R AU (S /\\ -R)) \\/ AG -R) AW R;",
"\nop leadstoAfter P S Q = -Q AW (Q /\\ AG (P => AF S));",
"\nop leadstoBetween P S Q R = AG (Q /\\ -R => ((P => -R AU (S /\\ -R)) \\/ AG - R) AW R);",
"\nop leadstoUntil P S Q R = AG (Q /\\ -R => (P => -R AU (S /\\ -R)) AW R);",
"\n",
"\n"
];


(* prints tl formulas by Prn *)
structure TtsLtlEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun printValue v =
    case v of
	VarExp (v,_) => print v
      | IntExp (i,_) => print (Int.toString i)
      | BoolExp b => print (if b then "1" else "0")
      | UnopExp (NOT,_,e) => (print "(- "; printValue e; print ")")
      | InfixExp (f,_,e1,e2) =>
	(case f of
	     AND => (print "("; printValue e1; print " /\\ "; printValue e2; print ")")
	   | OR  => (print "("; printValue e1; print " \\/ "; printValue e2; print ")")
	   | EQ  => (print "("; printValue e1; print " = "; printValue e2; print ")")
	   | NE  => (print "- ("; printValue e1; print " = "; printValue e2; print ")")
	   | LT  => (print "("; printValue e1; print " lt "; printValue e2; print ")")
	   | LE  => (print "("; printValue e1; print " le "; printValue e2; print ")")
	   | GT  => (print "("; printValue e1; print " gt "; printValue e2; print ")")
	   | GE  => (print "("; printValue e1; print " ge "; printValue e2; print ")")
	   | ADD => (print "("; printValue e1; print " + "; printValue e2; print ")")
	   | MUL => (print "("; printValue e1; print " * "; printValue e2; print ")"))
      | _ => print "?";  (* should not happen *)

fun printProbe pb =
    case pb of
        StateProbe (_,s) => print s
      | ValueProbe (_,e) => printValue e
      | EventProbe (_,l) => print l
      | TagProbe (_,l)   => print l
      | NotProbe r => (print "("; print "- "; printProbe r; print ")")
      | AndProbe (r1,r2) => (print "("; printProbe r1; print " /\\ "; printProbe r2; print ")")
      | OrProbe (r1,r2) => (print "("; printProbe r1; print " \\/ "; printProbe r2; print ")")
      | ImpProbe (r1,r2) => (print "("; printProbe r1; print " => "; printProbe r2; print ")")
      | _ => print "?";  (* should not happen *)


fun printProp p =
    case p of
        DeadlockFreeProp => print "deadlockFree"
      | InfinitelyOftenProp p => (print "infinitelyOften "; printProbe p)
      | MortalProp p => (print "mortal "; printProbe p)
      | LiveProp p => (print "live "; printProbe p)
      | PresentProp (p,NONE) => (print "present "; printProbe p)
      | PresentAfterProp (p1,p2,NONE) => (print "presentAfter "; printProbe p1; print " "; printProbe p2)
      | PresentBeforeProp (p1,p2,NONE) => (print "presentBefore "; printProbe p1; print " "; printProbe p2)
      | PresentBetweenProp (p1,p2,p3,NONE) => (print "presentBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PresentUntilProp (p1,p2,p3,NONE) => (print "presentUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AbsentProp (p,NONE) => (print "absent "; printProbe p)
      | AbsentAfterProp (p1,p2,NONE) => (print "absentAfter "; printProbe p1; print " "; printProbe p2)
      | AbsentBeforeProp (p1,p2,NONE) => (print "absentBefore "; printProbe p1; print " "; printProbe p2)
      | AbsentBetweenProp (p1,p2,p3,NONE) => (print "absentBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AbsentUntilProp (p1,p2,p3,NONE) => (print "absentUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoProp (p1,p2,NONE) => (print "leadsto "; printProbe p1; print " "; printProbe p2)
      | LeadstoBeforeProp (p1,p2,p3,NONE) => (print "leadstoBefore "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoAfterProp (p1,p2,p3,NONE) =>  (print "leadstoAfter "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoBetweenProp (p1,p2,p3,p4,NONE) => (print "leadstoBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | LeadstoUntilProp (p1,p2,p3,p4,NONE) => (print "leadstoUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | AlwaysProp (p,NONE) => (print "always "; printProbe p)
      | AlwaysAfterProp (p1,p2,NONE) => (print "alwaysAfter "; printProbe p1; print " "; printProbe p2)
      | AlwaysBeforeProp (p1,p2,NONE) => (print "alwaysBefore "; printProbe p1; print " "; printProbe p2)
      | AlwaysBetweenProp (p1,p2,p3,NONE) => (print "alwaysBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AlwaysUntilProp (p1,p2,p3,NONE) => (print "alwaysUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesProp (p1,p2,NONE) => (print "precedes "; printProbe p1; print " "; printProbe p2)
      | PrecedesBeforeProp (p1,p2,p3,NONE) => (print "precedesBefore "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesAfterProp (p1,p2,p3,NONE) => (print "precedesAfter "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesBetweenProp (p1,p2,p3,p4,NONE) => (print "precedesBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | PrecedesUntilProp (p1,p2,p3,p4,NONE) => (print "precedesUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | AndProp (p1,p2) => (print "("; printProp p1; print " /\\ "; printProp p2; print ")")
      | OrProp (NotProp p1,p2) => (print "("; printProp p1; print " => "; printProp p2; print ")")
      | OrProp (p1,p2) => (print "("; printProp p1; print " \\/ "; printProp p2; print ")")
      | ImpProp (p1,p2) => (print "("; printProp p1; print " => "; printProp p2; print ")")
      | NotProp p => (print "- "; print "("; printProp p; print ")")
      | BoxProp p => (print "[] "; print "("; printProp p; print ")")
      | DiaProp p => (print "<> "; print "("; printProp p; print ")")
      | NextProp p => (print "() "; print "("; printProp p; print ")")
      | UntilProp (p1,p2) => (print "("; printProp p1; print " U "; printProp p2; print ")")
      | ReleaseProp (p1,p2) => (print "("; printProp p1; print " V "; printProp p2; print ")")
      | ProbeProp p => (print "("; printProbe p; print ")")
      | VarProp n => print n
      | _ => (* timed property left *) print "?timed?";

fun printProperty (PropertyDecl (s,p)) = (newline 0; print "op "; print s; print " = "; printProp p; print ";")
  |  printProperty (Assert ("",_,_)) = () (* should not happen *)
  |  printProperty (Assert (s,tt,ff)) =
     let fun pack (SOME s) df = (case String.sub (s,0) of #"\"" => s | _ => ("\"" ^ s ^ "\"")) | pack _ df = df
     in case (pack tt "\"TRUE\"",pack ff "\"FALSE\"") of
	    ("\"TRUE\"","\"FALSE\"") => (newline 0; print s; print ";")
	  | (ptt,pff) => (newline 0; print "assert "; print s; print " "; print ptt; print " "; print pff; print ";")
     end;

fun emitFiacre (Program (_::[(_,dl,_)])) =
    ((* insert patterns.ltl library into file *)
     print (case !LOGIC of "selt" => ltllib | "muse" => mmclib | _ => failwith "unknown logic");
     case !DEADTAGS of
	 [] => ()
       | l => (newline 0; newline 0; print "# dead transitions possibly referred to in user formulas";
	       List.app (fn g => (newline 0; print "op "; print g; print " = F;")) l);
     case !JPS of
	 [] => ()
       | h::t => (newline 0; newline 0; print "# transitions with positive eft";
		  newline 0; 
		  print "op timed_events = "; print h; List.app (fn g => (print " \\/ "; print g)) t;
		  print ";");
     case dl of
	 [] => ()
       | _ => (newline 0; newline 0; print "# user formulas";
	       List.app printProperty dl);
     newline 0; newline 0)
    handle BYE => raise BYE | _ => failwith "emit tl formulas";

end; (* structure TtsLtlEmit *)


	  

(* prints the TTS Net by Prn *)
structure TtsNetEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun print_Fiacre (Program ((_,_,ProcessDecl d)::_)) =
        print_Process 0 d
					       
and print_Process i (a,_,_,b,_,_,_,_,_,StTo e,c) =
	(newline i; print "net process_"; print a;
         newline i; newline i; print_StatesDecl (b,e);
	 app (print_Transition i) c;
	 app (print_prior i) c;
	 newline i)

(* each state in initial state is a marked place *)
(* ##### also prints other states in case no transition refers to it. wise ? ##### *)
and print_StatesDecl (a,b) =
    List.app (fn s => (print "pl "; print s; if mem s b then print " (1)" else print " (0)"; print "\n")) a

and print_Interval (left,right) =
    let fun printb b =
	    print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
	    handle Overflow => emit_error "interval bound too large"
	val Z = ref false
    in case left of
	   Open b => (print " ]"; printb b)
	 | Closed b => (Z := b <= 0.0; print " ["; printb b)
	 | Infinite => print " ]-w";
       print ",";
       case right of
	   Open b => (printb b; print "[")
	 | Closed b => (printb b; print "]")
	 | Infinite => if !AADL andalso !Z then print "0]" else print "w["
    end

and positiveEft (left,_) =
    case left of Closed b => b > 0.0 | Infinite => false | _ => true

and print_Transition i (([tag],lab,_),src,c,interv,_) =
    let val tgt = get_target c
	val (read,inhib,from,tg) =  (* inhib for probes, in case needed *)
	    if !AADL then
		let (* handles inh and loops in fiacre source *)
		    val (rd,fr) = List.partition (fn p => mem p tgt) src
		    val (rfr,fr) = List.partition (fn p => String.sub(p,String.size p - 1) = #"?") fr
		    val (inh,fr) = List.partition (fn p => String.sub(p,String.size p - 2) = #"?") fr  (* ends with "?-" *)
		in (map (fn s => s ^ "?") rd @ rfr,
		    inh,
		    fr,
		    List.filter (fn p => not (mem p rd)) tgt)
		end
	    else
		let val (rd,src') = List.partition (fn p => String.sub(p,String.size p - 1) = #"?") src
		    val (inh,fr) = List.partition (fn p => String.sub(p,String.size p - 2) = #"?") src'  (* ends with "?-" *)
		in (rd,
		    inh,
		    fr,
		    tgt)
		end
    in newline i; print "tr "; print tag; print " "; case lab of "" => () | _ => (print ": "; print lab);
       print_Interval interv;
       if !EMITPOS andalso positiveEft interv then JPS := tag :: !JPS else ();
       print_list (" "," ","") (fn p => print (p ^ "1")) read;
       print_list (" "," ","") (fn p => print (p ^ "1")) inhib;
       print_list (" "," ","") print from;
       print " -> ";
       print_list (""," ","") print tg
    end

(* gets target state *)
and get_target st =
    case st of
	StSequence(stm1,stm2) => get_target stm1 @ get_target stm2
      | StTo stl => stl
      | _ => []

and print_prior i (_,_,_,_,[]) = ()
  | print_prior i (([tg],_,_),_,_,_,prl) = (newline i; print "pr "; print tg; print " > "; print_list (""," ","") print prl)

fun emitFiacre s0 = (print_Fiacre s0; newline 0; flush ()) handle BYE => raise BYE | _ => failwith "emit net";

end; (* structure TtsNetEmit *)




(* top routine *)
	  
structure EmitTts : EMIT = struct

fun readOnly f = ();
    (*
     let open Posix.FileSys
	 val m = S.flags [S.irusr, S.irgrp, S.iroth]
     in chmod (f, m)
     end;
     *)

fun emit t =
    (* ### ALL files in tts directory should be read-only (A FAIRE) ### *)
    let val dir = !OutPrint.name
	val base = #base(OS.Path.splitBaseExt (#file (OS.Path.splitDirFile dir)))
	val fnet = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "net"}}
	val snet = TextIO.openOut fnet handle x => emit_error ("cannot open "^fnet)
	val fdat = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "c"}}
	val sdat = TextIO.openOut fdat handle x => emit_error ("cannot open "^fdat)
	val fdef = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "h"}}
	val sdef = TextIO.openOut fdef handle x => emit_error ("cannot open "^fdef)
    in OutPrint.name := fnet;
       OutPrint.FILE := snet;
       printVerbose "::Writing Net(.net) File\n";
       TtsNetEmit.emitFiacre t;
       readOnly fnet;
       printVerbose "::Done\n";
       let (* ugly but will improve later *)
	   fun writeh f x =
	       (OutPrint.flush (); OutPrint.name := fdef; OutPrint.FILE := sdef;
		f x;
		OutPrint.flush (); OutPrint.name := fdat; OutPrint.FILE := sdat)
       in OutPrint.name := fdat;
	  OutPrint.FILE := sdat;
	  readOnly fdat;
	  printVerbose "::Writing Data(.h|.c) Files\n";		 
	  TtsCEmit.emitFiacre (base^".h",fdef,writeh) t;
	  readOnly fdef;
	  printVerbose "::Done\n"
       end;
       if !WITHJSON then
	   ((* copy input file into tts dir *)
	    let val infile = !Input.name
		val outfile = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "fcr"}}
	    in copyFile (infile,outfile);
	       readOnly outfile
	    end;
	    (* create json description file into tts dir *)
	    let val fjson = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "json"}}
		val sjson = TextIO.openOut fjson handle x => emit_error ("cannot open "^fjson)
	    in OutPrint.name := fjson;
	       OutPrint.FILE := sjson;
	       Json.printJson ();
	       readOnly fjson
	    end)
       else
	   ();
       if !PROPERTIES then
	   let val ext = case !LOGIC of "selt" => "ltl" | "muse" => "mmc" | _ => "?" (* should not happen *)
	       val fltl = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME ext}}
	       val sltl = TextIO.openOut fltl handle x => (ErrPrint.print ("cannot open "^fltl^"\n"); raise BYE)
	   in OutPrint.name := fltl;
	      OutPrint.FILE := sltl;
	      printVerbose ("::Writing Ltl (." ^ ext ^ ") Files\n");		 
	      TtsLtlEmit.emitFiacre t;
	      readOnly fltl;
	      printVerbose "::Done\n"
	   end
       else
	   ()
    end;

end;


